;===============================================================================
;===                            PUD file I/O                                 ===
;===============================================================================




; Some arbitrary reality check values so we don't try to load garbage

MIN_PUD_FILE  equ 100h    ; minimum file size         ( 256B  )
MAX_PUD_FILE  equ 500000h ; maximum file size          (  5MB )
MIN_SECTION   equ 0Ch     ; minimum total section size ( 12B  )
MAX_SECTION   equ 100000h ; maximum section data size  (  1MB )

;
; |=- PUD file format -=|
;
; ( pretty similar to RIFF )
;
;   {   4 bytes - 4CC section name
;       4 bytes - DWORD section length
;       n bytes - section data          }  repeat until EOF
;


.data?
  NumberOfBytesRead   dd ?
  save_fail           dd ?

.code

open_file_read proc path:DWORD
  ; open a file for reading
  push 0
  push 0
  push OPEN_EXISTING
  push 0
  push FILE_SHARE_READ
  push GENERIC_READ
  push path
  call CreateFileA
  ret
open_file_read endp


open_file_write proc path:DWORD
  ; open a file for writing
  push 0
  push FILE_ATTRIBUTE_NORMAL
  push CREATE_ALWAYS
  push 0
  push FILE_SHARE_WRITE
  push GENERIC_WRITE
  push path
  call CreateFileA
  ret
open_file_write endp

read_section proc ptrSec:DWORD,ptrData:DWORD,eof:DWORD
  ; allocate memory for section data and read it from the memory mapped file.
  ; fill the details in the passed section header
  ;
  ; return ptr to the next unread byte or 0 on fail
  
  
  xor eax,eax       ; zero return value
  
  mov esi,ptrData   ; esi = current read pointer
  
  mov edx,eof
  sub edx,esi       ; edx = file bytes remaining
  
  mov ebx,[esi]     ; ebx = section name
  mov ecx,[esi+4]   ; ecx = section data size
  add esi,8
    
  ; section size invalid?
  cmp ecx,MAX_SECTION
  ja @F
    
    ; section ends beyond eof?
    cmp edx,ecx
    jb @F
      
      ; allocate section data buffer
      push ecx
      call doAlloc
  
      ; buffer allocated ok?
      test eax,eax
      jz @F
	
	; fill the section header
	mov edi,ptrSec
	mov [edi+SECTION_NAME],ebx
	mov [edi+SECTION_SIZE],ecx
	mov [edi+SECTION_DATA],eax
	  
	;copy the data
	
	mov edi,eax ; destination is newly allocated buffer
	
	; "movsb" is one of the "string instructions"
	; these are the only instructions that can be 
	; used with the "rep" prefix (also "repz" and "repnz")
	;
	; "movs" stands for "move string" the 'b'=BYTE
	;  ( there is also "movsw" and "movsd" )
	;
	; the "movsb" instruction tells the CPU to:
	;
	;   mov a byte from [esi] to [edi]
	;   inc esi
	;   inc edi
	;   dec ecx
	;
	; the "rep" prefix tells it to "repeat" until ecx=0
	
	rep movsb
	
	;return ptr to next read pos
	mov eax,esi
  @@:
  ret
read_section endp



read_sections proc buff:DWORD, eof:DWORD
  ; read each section memory mapped PUD
  ; return handle to a new pud object
  ;
  ; buff = address of file map
  ; eof  = address of the end of the file
  
  
  LOCAL hPud
  LOCAL hSection
  
  ; allocate the first section header (this proc is in PUDpud.inc)
  call new_section
  
  mov hPud,eax    ; the first section handle is the pud handle
  mov hSection,eax

  ; section hdr allocated ok?
  test eax,eax
  jz exit_read_sections
    
    ;start reading at start of file buffer
    mov esi,buff
    
    section_loop:
      
      push eof      ; ptr to EOF
      push esi      ; current read ptr
      push eax      ; ptr to new section header
      call read_section
      
      ; section read ok?
      test eax,eax
      jz section_bad_read
      
      ;update the source ptr
      mov esi,eax
      
      ; enough file left to read another section?
      mov edi,esi
      sub edi,eof         ; edi=number of bytes left in file
      cmp edi,MIN_SECTION
      jb exit_read_sections
	
	; get a new section header
	call new_section
	
	; new hdr allocated ok?
	test eax,eax
	jz section_bad_read
	  
	  ;link it to the list
	  mov edi,hSection
	  mov [edi],eax
	  
	  ;save the new section handle for the next link
	  mov hSection,eax
	  
	  ;read the next section
	  jmp section_loop
  
  section_bad_read:
  ; file is invalid, or cant be allocated
  ; some sections may have been allocated
  ; free any allocated memory
  push hPud
  call PUDdestroy
  mov hPud,0
    
  exit_read_sections:
  mov eax,hPud
  ret
read_sections endp

flen proc hFile:DWORD
  ; return file size if valid or 0 on fail
  ;
  ; this function does some reality checking on the file
  ; size in case we accidentally try to open some huge 
  ; or tiny file
  
  push 0
  push hFile
  call GetFileSize
  cmp eax,INVALID_FILE_SIZE
  jnz @F
    cmp eax,MAX_PUD_FILE
    jbe @F
      cmp eax,MIN_PUD_FILE
      ja @F
	xor eax,eax
  @@:
  ret
flen endp


read_pud_file proc hFile:DWORD
  ; Here it is quick and easy to just map the PUD file into memory.
  ;
  ; From there we call "read_sections" to read the individual 
  ; sections from the memory-mapped file.
  
  LOCAL hMap      ; handle to the file mapping object
  LOCAL ptrMap    ; pointer to the memory mapped file
  LOCAL file_size ; size in bytes of the PUD file
  LOCAL hPud      ; handle to the read pud object (return value)
  
  ; LOCALs are not guaranteed to be zeroed when created
  xor eax,eax
  mov hPud,eax  ; zero pud handle
  
  ; get file size
  push hFile
  call flen
  
  ; valid file size?
  test eax,eax
  jz exit_pud_read
    
    mov file_size,eax
    
    ; create file mapping object
    push 0
    push 0
    push 0
    push PAGE_READONLY
    push 0
    push hFile
    call CreateFileMapping
    mov hMap,eax
    test eax,eax
    
    ;created ok?
    jz exit_pud_read
      
      ; map the file to memory
      push 0
      push 0
      push 0
      push FILE_MAP_READ
      push eax
      call MapViewOfFile
      mov ptrMap,eax
      test eax,eax
      
      ;mapped ok?
      jz close_mapping
	
	; calc eof address
	add eax,file_size   
	
	; read sections
	push eax
	push ptrMap
	call read_sections
	mov hPud,eax
	
	; unmap the file
	push ptrMap
	call UnmapViewOfFile
      
      close_mapping:
      ; close the file mapping object
      push hMap
      call CloseHandle
	
  exit_pud_read:
  
  ; return handle to pud object (if we got one)
  mov eax,hPud
  ret
read_pud_file endp



PUDload proc path:DWORD
  ; open a pud file and pass file handle to read proc
  ; or display an error message if open fails
  
  ; return handle to pud object or 0 on fail
  
  LOCAL hFile
  
  push path
  call open_file_read
  
  
  cmp eax,INVALID_HANDLE_VALUE
  
  je load_open_error
    ;store the file handle
    mov hFile,eax
    
    ; call the read proc
    push eax
    call read_pud_file
    
    push eax
      
      ; close the file
      push hFile
      call CloseHandle
    
    pop eax
    ret
  
  load_open_error:
  push 0
  push OFFSET AppName
  push OFFSET szCantOpen
  push hWindow
  call MessageBoxA
  xor eax,eax
  ret
PUDload endp


write_pud_section proc hSection:DWORD, hFile:DWORD
  ; Write a pud section to file hFile
  ;
  ; No need to mess around with buffers here, we just
  ; dump the section name, size then the data straight
  ; to the file.
  mov eax,hSection
  
  push eax
    ; write section name and size
    push 0
    push OFFSET NumberOfBytesRead
    push 8
    lea ebx,[eax+4]
    
    push ebx
    push hFile
    call WriteFile
    test eax,eax
    jnz @F
      inc save_fail
      popad
      ret
    @@:
  pop eax
  
  ; write section data
  push 0
  push OFFSET NumberOfBytesRead
  push [eax+SECTION_SIZE]
  push [eax+SECTION_DATA]
  push hFile
  call WriteFile
  test eax,eax
  jnz @F
    inc save_fail
  @@:
  ret
write_pud_section endp

write_pud_file proc hPud:DWORD, hFile:DWORD
  ; writing is fairly straight-forward, we just
  ; walk through our linked list of headers and
  ; write each section one at a time.
  
  mov save_fail,0
  
  ; I ended up using a global flag to signal write
  ; failures, just because I'm lazy and it seemed
  ; easier at the time.
  
  mov eax,hPud
  @@:
    push eax
      push hFile
      push eax
      call write_pud_section
    pop eax
    mov eax,[eax]
    test eax,eax
  jnz @B
  ret
write_pud_file endp


PUDsave proc hPud:DWORD,path:DWORD
  ; Save pud object to path
  ; return 0 on fail
  ;
  ; here we just open the output file for writing
  ; then pass the handles to "write_pud_file"
  LOCAL hFile
  
  ; do we have a pud object to write?
  cmp hPud,0
  je @F
  
    ; open the output file
    push path
    call open_file_write
    
    ; opened ok?
    cmp eax,INVALID_HANDLE_VALUE
    je @F
      mov hFile,eax

      ; write the PUD sections
      push eax
      push hPud
      call write_pud_file
  
      ; close the file
      push hFile
      call CloseHandle
      
      ; did we have any write issues?
      cmp save_fail,0
      jne @F
	
	; PUD file saved successfully
	; ( unless CloseHandle failed in
	;   which case eax will be 0 - fail )
	ret
      
  @@:
  ; no pud, file open failure 
  ; or write fail flag is set
  ;
  ; return zero 
  xor eax,eax
  ret
PUDsave endp

;
; TO DO here... maybe 
;
; check if the output file exists, then rename it to a temp name before
; overwriting, so it can be restored if the write fails for some reason
;
; better error reporting, not just "it worked" or "it didn't"
;


















